/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "stdafx.h"
#include "IFOUtils.h"
#include "Utils.h"
#include "Cell.h"

char *IFONames[100] = {"VIDEO_TS.IFO","VTS_01_0.IFO","VTS_02_0.IFO","VTS_03_0.IFO","VTS_04_0.IFO","VTS_05_0.IFO","VTS_06_0.IFO","VTS_07_0.IFO","VTS_08_0.IFO","VTS_09_0.IFO","VTS_10_0.IFO","VTS_11_0.IFO","VTS_12_0.IFO","VTS_13_0.IFO","VTS_14_0.IFO","VTS_15_0.IFO","VTS_16_0.IFO","VTS_17_0.IFO","VTS_18_0.IFO","VTS_19_0.IFO","VTS_20_0.IFO","VTS_21_0.IFO","VTS_22_0.IFO","VTS_23_0.IFO","VTS_24_0.IFO","VTS_25_0.IFO","VTS_26_0.IFO","VTS_27_0.IFO","VTS_28_0.IFO","VTS_29_0.IFO","VTS_30_0.IFO","VTS_31_0.IFO","VTS_32_0.IFO","VTS_33_0.IFO","VTS_34_0.IFO","VTS_35_0.IFO","VTS_36_0.IFO","VTS_37_0.IFO","VTS_38_0.IFO","VTS_39_0.IFO","VTS_40_0.IFO","VTS_41_0.IFO","VTS_42_0.IFO","VTS_43_0.IFO","VTS_44_0.IFO","VTS_45_0.IFO","VTS_46_0.IFO","VTS_47_0.IFO","VTS_48_0.IFO","VTS_49_0.IFO","VTS_50_0.IFO","VTS_51_0.IFO","VTS_52_0.IFO","VTS_53_0.IFO","VTS_54_0.IFO","VTS_55_0.IFO","VTS_56_0.IFO","VTS_57_0.IFO","VTS_58_0.IFO","VTS_59_0.IFO","VTS_60_0.IFO","VTS_61_0.IFO","VTS_62_0.IFO","VTS_63_0.IFO","VTS_64_0.IFO","VTS_65_0.IFO","VTS_66_0.IFO","VTS_67_0.IFO","VTS_68_0.IFO","VTS_69_0.IFO","VTS_70_0.IFO","VTS_71_0.IFO","VTS_72_0.IFO","VTS_73_0.IFO","VTS_74_0.IFO","VTS_75_0.IFO","VTS_76_0.IFO","VTS_77_0.IFO","VTS_78_0.IFO","VTS_79_0.IFO","VTS_80_0.IFO","VTS_81_0.IFO","VTS_82_0.IFO","VTS_83_0.IFO","VTS_84_0.IFO","VTS_85_0.IFO","VTS_86_0.IFO","VTS_87_0.IFO","VTS_88_0.IFO","VTS_89_0.IFO","VTS_90_0.IFO","VTS_91_0.IFO","VTS_92_0.IFO","VTS_93_0.IFO","VTS_94_0.IFO","VTS_95_0.IFO","VTS_96_0.IFO","VTS_97_0.IFO","VTS_98_0.IFO","VTS_99_0.IFO"};

#define MAX_IFO_SIZE_INCREASE	100		// Maximum number of LBAs, the IFO size can be increased.
#define MAX_NUM_VOBUS 64000

CIFOLoader::CIFOLoader(int VTS)
{
	m_Buffer = NULL;
	m_IFOfd = NULL;
	m_NumVOBUs = 0;
	m_LoadedOK = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	LoadIFO(VTS);
}


CIFOLoader::CIFOLoader(const char* IFOFileName)
{
	m_Buffer = NULL;
	m_IFOfd = NULL;
	m_NumVOBUs = 0;
	m_LoadedOK = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	LoadIFO(IFOFileName);
}

CIFOLoader::CIFOLoader()
{
	m_IFOfd = NULL;
	m_IsManager = 0;
	m_Buffer = 0;
	strcpy(m_IFOFileName,"");
	m_IFOSizekB = 0;
	m_MaxIFOSize = 0;
	// JEAN: FIX THIS!
	m_NumVOBUs = 0;
	m_VOBUStartLBA = new uint[MAX_NUM_VOBUS];
	m_VTSNum = -1;
}

int CIFOLoader::LoadIFO(const char* IFOFileName)
{
	m_LoadedOK = false;
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = fopen(IFOFileName,"rb+");
	// If you can't read/write, let's see if we can at least read.
	if(m_IFOfd == 0)
		m_IFOfd = fopen(IFOFileName,"rb");
	m_IsManager = strcmpi(GetFname(IFOFileName),"video_ts") == 0;
	delete[] m_Buffer;
	m_Buffer = NULL;
	strcpy(m_IFOFileName,IFOFileName);
	m_IFOSizekB = GetFileSizekB(IFOFileName);
	m_MaxIFOSize = m_IFOSizekB*1024 + MAX_IFO_SIZE_INCREASE*SECTOR_SIZE;
	if(m_IFOSizekB > 0)
		m_Buffer = new uchar[m_MaxIFOSize];
	if(m_IFOfd == NULL)
		return -1;
	fread(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd);
	m_NumVOBUs = 0;
	int i;
	// Find which VTS this is...
	char* Temp = GetFnameExt(IFOFileName);
	for(i=0;i<100;i++)
	{
		if(strcmpi(Temp,IFONames[i]) == 0)
		{
			m_VTSNum = i;
			break;
		}
	}
	m_LoadedOK = true;
	return 0;
}	

int CIFOLoader::LoadIFO(int VTS)
{
	m_LoadedOK = false;
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = fopen(IFONames[VTS],"rb+");
	// If you can't read/write, let's see if we can at least read.
	if(m_IFOfd == 0)
		m_IFOfd = fopen(IFONames[VTS],"rb");
	m_IsManager = VTS == 0;
	m_VTSNum = VTS;
	delete[] m_Buffer;
	m_Buffer = NULL;
	strcpy(m_IFOFileName,IFONames[VTS]);
	m_IFOSizekB = GetFileSizekB(IFONames[VTS]);
	m_MaxIFOSize = m_IFOSizekB*1024 + MAX_IFO_SIZE_INCREASE*SECTOR_SIZE;
	if(m_IFOSizekB > 0)
		m_Buffer = new uchar[m_MaxIFOSize];
	if(m_IFOfd == NULL)
		return -1;
	fread(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd);
	m_NumVOBUs = 0;
	m_LoadedOK = true;
	return 0;
}

void CIFOLoader::CloseIFO()
{
	if(m_IFOfd)
		fclose(m_IFOfd);
	m_IFOfd = 0;
}


CIFOLoader::~CIFOLoader()
{
	if(m_IFOfd) fclose(m_IFOfd);
	delete[] m_Buffer;
	delete[] m_VOBUStartLBA;
}

uchar* CIFOLoader::FindTable(IFOTable Table)
{
	// Load main table.
	IFO_main* Main = (IFO_main*) m_Buffer;

	uint Offset;

	switch(Table)
	{
	case VTSM_PGCI_UT :		Offset = (m_IsManager) ? GetInt(m_Buffer,0x00c8) :  GetInt(m_Buffer,0x00d0) ; break;
	case VTSM_C_ADT :		Offset = GetInt((uchar*)&Main->m_PVTSM_C_ADT,0) ; break;
	case VTSM_VOBU_ADMAP:	Offset = GetInt((uchar*)&Main->m_PVTSM_VOBU_ADMAP,0) ; break;
	case TT_SRPT:			assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PTT_SRPT,0) ; break;
	case VTS_PTT_SRPT:		assert(!m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_PTT_SRPT,0) ; break;
	case VMGM_PGCI_UT:		assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_PTT_SRPT,0) ; break;
	case VTS_C_ADT :		Offset = GetInt((uchar*)&Main->m_PVTS_C_ADT,0) ; break;
	case VTS_PGCI :			Offset = GetInt((uchar*)&Main->m_PVTS_PGCI,0) ; break;
	case VTS_VOBU_ADMAP:	Offset = GetInt((uchar*)&Main->m_PVTS_VOBU_ADMAP,0) ; break;
	case VTS_TMAPTI:		assert(!m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_TMAPTI,0) ; break;
	case VMG_TXTDT_MG:		assert(m_IsManager) ; Offset = GetInt((uchar*)&Main->m_PVTS_TMAPTI,0) ; break;
	default: assert(0); return 0;
	}
	if(Offset == 0)	// No table.
		return 0;
	return m_Buffer+Offset*SECTOR_SIZE;
}

uint CIFOLoader::GetEndAddress(IFOTable Table)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr == 0)
		return 0;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP)
		return GetInt(Ptr,0x04);
	else
		return GetInt(Ptr,0);
}

uint CIFOLoader::GetAbsoluteEndAddress(IFOTable Table)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr == 0)
		return 0;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP)
		return Ptr-m_Buffer+GetInt(Ptr,0x04);
	else
		return Ptr-m_Buffer+GetInt(Ptr,0);
}

void CIFOLoader::IncEndAddress(IFOTable Table, int IncBytes)
{
	uchar *Ptr = FindTable(Table);
	if(Ptr == 0) return;
	if(Table != VTSM_VOBU_ADMAP && Table != VTS_VOBU_ADMAP)
		AddInt(Ptr,0x04,IncBytes);
	else
		AddInt(Ptr,0,IncBytes);
}

bool CIFOLoader::TableSizeCanChange(IFOTable Table, int NumBytes)
{
	// Load main table.
	IFO_main* Main = (IFO_main*) m_Buffer;

	if(Table == VTSM_PGCI_UT || Table == VTSM_C_ADT || Table == VTSM_VOBU_ADMAP || 
		Table == VTS_PGCI || Table == VTS_C_ADT || Table == VTS_VOBU_ADMAP ||
		Table == VTS_TMAPTI)
	{
		uint EndAddress = GetEndAddress(Table);
		if(EndAddress == 0)
			return false;
		// We can change the size if the incremented last byte is in the same sector.
		return ((EndAddress + NumBytes) / SECTOR_SIZE == EndAddress / SECTOR_SIZE);
	}
	assert(0);
	return 0;
}

void CIFOLoader::ChangeTableSize(IFOTable Table, int NumBytes)
{
	if(TableSizeCanChange(Table, NumBytes))
		return;

	uchar *Ptr = FindTable(Table);
	assert(Ptr);	// JUST SO I don't forget to check!
	if(Ptr==0)return;

	// We need to move stuff up or down!
	uint EndAddress = GetEndAddress(Table);
	uchar* CurrentEndLBAPtr = Ptr + (EndAddress / SECTOR_SIZE + 1)*SECTOR_SIZE;
	uchar* DestEndLBAPtr = Ptr + ((EndAddress+NumBytes) / SECTOR_SIZE + 1)*SECTOR_SIZE;
	int Inc = (int)DestEndLBAPtr - (int)CurrentEndLBAPtr;	// This is by how much we move.
	memmove(DestEndLBAPtr,CurrentEndLBAPtr,m_IFOSizekB*1024-(CurrentEndLBAPtr-m_Buffer));
	Inc /= SECTOR_SIZE;

	// Now update any pointer that might point past NextLBAPtr.
	IFO_main* Main = (IFO_main*) m_Buffer;
	uint ThisTableSector = (Ptr - m_Buffer) / SECTOR_SIZE;
	if(m_IsManager)
	{
		if(FixEndian(Main->m_PTT_SRPT) > ThisTableSector)
			AddInt((uchar*)&Main->m_PTT_SRPT,0,Inc);
	}

	if(FixEndian(Main->m_PVTS_PTT_SRPT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_PTT_SRPT,0,Inc);

	if(FixEndian(Main->m_PVTS_PGCI) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_PGCI,0,Inc);

	if(FixEndian(Main->m_PVTSM_PGCI_UT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_PGCI_UT,0,Inc);

	if(FixEndian(Main->m_PVTS_TMAPTI) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_TMAPTI,0,Inc);

	if(FixEndian(Main->m_PVTSM_C_ADT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_C_ADT,0,Inc);

	if(FixEndian(Main->m_PVTSM_VOBU_ADMAP) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTSM_VOBU_ADMAP,0,Inc);

	if(FixEndian(Main->m_PVTS_C_ADT) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_C_ADT,0,Inc);

	if(FixEndian(Main->m_PVTS_VOBU_ADMAP) > ThisTableSector)
		AddInt((uchar*)&Main->m_PVTS_VOBU_ADMAP,0,Inc);
	
	// Also, modify last sector of IFO.
	AddInt((uchar*)&Main->m_IFOLastSector,Inc,0);

	// Finally, reflect change in size of file.
	m_IFOSizekB += Inc*SECTOR_SIZE/1024;

}

void CIFOLoader::MoveDataInTable(IFOTable Table, uchar* Ptr, int NumBytes)
{
	if(NumBytes >= 0)
	{
		// We're growing, change table size first!
		ChangeTableSize(Table, NumBytes);
		uint EndAddress = GetEndAddress(Table);
		// Then move data. FindTable(Table) + GetEndAddress(Table) point to the last element of the table.
		int NumMove = FindTable(Table) + GetEndAddress(Table) - Ptr + 1;
		memmove(Ptr+NumBytes, Ptr, NumMove);
		IncEndAddress(Table,NumBytes);
	}
	else
	{
		// We're shrinking! Move things first, then change table size.
		uint EndAddress = GetEndAddress(Table);
		// Then move data. FindTable(Table) + GetEndAddress(Table) point to the last element of the table.
		int NumMove = FindTable(Table) + GetEndAddress(Table) - Ptr + 1;
		memmove(Ptr+NumBytes, Ptr, NumMove);
		ChangeTableSize(Table, NumBytes);
		IncEndAddress(Table,NumBytes);
	}
}

uchar* CIFOLoader::GoToLU(ushort LU)
{
	uchar*	Ptr = FindTable(VTSM_PGCI_UT);
	// Find offset to VTSM_LU
	return Ptr?Ptr+GetInt(Ptr,12+(LU-1)*8) : 0;
}	

uchar* CIFOLoader::GoToPCG(ushort LU, ushort PGC)
{
	if(LU!=0)
	{
		uchar*	Ptr = GoToLU(LU);
		return Ptr ? Ptr + GetInt(Ptr,12+(PGC-1)*8) : 0;
	}
	else
	{
		uchar*	Ptr = FindTable(VTS_PGCI);
		return Ptr ? Ptr + GetInt(Ptr,12+(PGC-1)*8) : 0;
	}
}
	
ushort CIFOLoader::GetNumLU()
{
	uchar*	Ptr = FindTable(VTSM_PGCI_UT);
	return Ptr ? GetShort(Ptr,0) : 0;
}	

ushort CIFOLoader::GetNumPGC(ushort LU)
{
	if(LU!=0)
	{
		uchar*	Ptr = GoToLU(LU);
		// Go there and get the number of PGCs.
		return Ptr ? GetShort(Ptr,0) : 0;
	}
	else
	{
		uchar*	Ptr = FindTable(VTS_PGCI);
		return Ptr ? GetShort(Ptr,0):0;
	}
}	

uchar CIFOLoader::GetNumCells(ushort LU, ushort PGC)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	return Ptr[3];
}

CELL_PLAYBACK_INFO *CIFOLoader::GetCellPlaybackInfo(ushort LU, ushort PGC, ushort Cell)
{
	if(GetNumCells(LU,PGC) < Cell)
		return 0;
	uchar* Ptr = GoToPCG(LU, PGC);
	// Now find offset to playback info for this cell.
	ushort Offset = GetShort(Ptr,0xe8);
	uchar* Ptr2 = Ptr + Offset + (Cell-1) * sizeof(CELL_PLAYBACK_INFO);
	return (CELL_PLAYBACK_INFO*) Ptr2;
}

CELL_POS_INFO *CIFOLoader::GetCellPosInfo(ushort LU, ushort PGC, ushort Cell)
{
	if(GetNumCells(LU,PGC) < Cell)
		return 0;
	uchar* Ptr = GoToPCG(LU, PGC);
	// Now find offset to playback info for this cell.
	ushort Offset = GetShort(Ptr,0xea);
	uchar* Ptr2 = Ptr + Offset + (Cell-1) * sizeof(CELL_POS_INFO);
	return (CELL_POS_INFO*) Ptr2;
}

int CIFOLoader::FindThisCell(ushort VOBID, uchar CELLID, ushort *LUOut, ushort *PGCOut, ushort *CELLOut)
{
	ushort NumPGC,NumCell;
	int Found = 0;
	if(LUOut != NULL)
	{
		ushort NumLU = GetNumLU();

		// Looping through all the LUs, all the PGCs and all the Cells, looking for the right VCID.
		ushort LU,PGC,CELL;
		for(LU=1;LU<=NumLU;LU++)
		{
			NumPGC = GetNumPGC(LU);
			for(PGC=1;PGC<=NumPGC;PGC++)
			{
				NumCell = GetNumCells(LU,PGC);
				for(CELL=1;CELL<=NumCell;CELL++)
				{
					CELL_POS_INFO *Temp = GetCellPosInfo(LU,PGC,CELL);
					if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
					{
						// Found it!
						LUOut[Found] = LU;
						PGCOut[Found] = PGC;
						CELLOut[Found++] = CELL;
					}
				}
			}
		}
	}
	else
	{
		ushort PGC,CELL;
		NumPGC = GetNumPGC(0);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			NumCell = GetNumCells(0,PGC);
			for(CELL=1;CELL<=NumCell;CELL++)
			{
				CELL_POS_INFO *Temp = GetCellPosInfo(0,PGC,CELL);
				if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
				{
					// Found it!
					PGCOut[Found] = PGC;
					CELLOut[Found++] = CELL;
				}
			}
		}
	}
	return Found;
}

bool CIFOLoader::IsReferenced(ushort VOBID, uchar CELLID, bool Menu)
{
	int NumLU = Menu?GetNumLU():1;
	int j;
	ushort PGC,CELL;
	// Go through all LUs.
	for(j=1;j<=NumLU;j++)
	{
		int LU = (Menu)?j:0;
		int NumPGC = GetNumPGC(LU);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			int NumCell = GetNumCells(LU,PGC);
			for(CELL=1;CELL<=NumCell;CELL++)
			{
				CELL_POS_INFO *Temp = GetCellPosInfo(LU,PGC,CELL);
				if(FixEndian(Temp->m_VOBID) == VOBID && Temp->m_CELLID == CELLID)
				{
					return true;
				}
			}
		}
	}
	return false;
}


void CIFOLoader::GetCLUT(uchar *CLUT, ushort LU, ushort PGC, bool RGB)
{
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return;
	int i;
	uchar* Ptr = (uchar*) Header->m_Palette;
	for(i=0;i<16;i++)
	{
		if(!RGB)
		{
			// Y Cr Cb
			CLUT[3*i] = Ptr[4*i+1];
			CLUT[3*i+1] = Ptr[4*i+2];
			CLUT[3*i+2] = Ptr[4*i+3];
		}
		else
		{
			// http://www.fourcc.org/fccyvrgb.php
#if 1
			float R = Ptr[4*i+1] + 1.402f*(Ptr[4*i+2]-128);
			CLUT[3*i] = (uchar) max(0,min(R,255));
			float G = Ptr[4*i+1] - 0.34414f*(Ptr[4*i+3]-128) - 0.71414f*(Ptr[4*i+2]-128);
			CLUT[3*i+1] = (uchar) max(0,min(G,255));
			float B = Ptr[4*i+1] + 1.772f*(Ptr[4*i+3]-128);
			CLUT[3*i+2] = (uchar) max(0,min(B,255));
#else
			float R = 1.164f*(Ptr[4*i+1] - 16) + 1.596f*(Ptr[4*i+2]-128);
			CLUT[3*i] = (uchar) max(0,min(R,255));
			float G = 1.164f*(Ptr[4*i+1] - 16) - 0.391f*(Ptr[4*i+3]-128) - 0.813f*(Ptr[4*i+2]-128);
			CLUT[3*i+1] = (uchar) max(0,min(G,255));
			float B = 1.164f*(Ptr[4*i+1] - 16) + 2.018f*(Ptr[4*i+3]-128);
			CLUT[3*i+2] = (uchar) max(0,min(B,255));
#endif
		}
	}
}

int CIFOLoader::GetSubpicLangages(char* Languages, bool Menu)
{
	int NumLang = 0;
	if(Menu || m_IsManager)
	{
		NumLang = GetShort(m_Buffer,0x154);
		Languages[0] = m_Buffer[0x158];
		Languages[1] = m_Buffer[0x159];
	}
	else
	{
		NumLang = GetShort(m_Buffer,0x254);
		int i;
		for(i=0;i<NumLang;i++)
		{
			Languages[2*i] = m_Buffer[0x258+i*6];
			Languages[2*i+1] = m_Buffer[0x259+i*6];
		}
	}
	return NumLang;
}

bool CIFOLoader::GetPgcSubpicLanguage(char* Languages, ushort LU, ushort PGC, uchar StreamID)
{
	char DLanguages[128];
	// First get the languages defined in the domain.
	int NumLang = GetSubpicLangages(DLanguages,LU!=0?true:false);
	// Now go the the PGC.
	PGC_HEADER *Header = GetPGCHeader(LU,PGC);
	assert(Header != 0);
	if(Header == 0)
		return 0;

	int i;
	for(i=0;i<32;i++)
	{
		int Attributes = Header->m_SubpicStreams[i];
		// Is the stream available?
		if(Attributes & 0x80000000 == 0)
			continue;
		// The stream number in Attributes is between 0 and 32, and corresponds to streams
		// 0x20 to 0x3F
		if(((Attributes>>24) & 0x1F) == StreamID-0x20) break;
		if(((Attributes>>16) & 0x1F) == StreamID-0x20) break;
		if(((Attributes>>8) & 0x1F) == StreamID-0x20) break;
		if(((Attributes) & 0x1F) == StreamID-0x20) break;
	}
	if(i==32)
	{
		Languages[0] = ' ';
		Languages[1] = ' ';
		return false;
	}
	Languages[0] = DLanguages[2*i];
	Languages[1] = DLanguages[2*i+1];
	if(Languages[0] == 0) Languages[0] = ' ';
	if(Languages[1] == 0) Languages[1] = ' ';
	return true;
}

PGC_HEADER* CIFOLoader::GetPGCHeader(ushort LU, ushort PGC)
{
	uchar* Ptr = GoToPCG(LU, PGC);
	return (PGC_HEADER*) Ptr;
}

void CIFOLoader::WriteIFO(char* WritePath)
{
	if(WritePath == NULL)
	{
		// Overwrite input IFO file.
		fseek(m_IFOfd,0,SEEK_SET);
		fwrite(m_Buffer,m_IFOSizekB*1024,1,m_IFOfd);
		fflush(m_IFOfd);
		FILE* fd = fopen(ChangeExt(m_IFOFileName,".BUP"),"wb");
		if(fd)
		{
			fwrite(m_Buffer,m_IFOSizekB*1024,1,fd);
			fclose(fd);
		}
	}
	else
	{
		FILE* fd = fopen(WritePath,"wb");
		if(fd)
		{
			fwrite(m_Buffer,m_IFOSizekB*1024,1,fd);
			fclose(fd);
		}
		fd = fopen(ChangeExt(WritePath,".BUP"),"wb");
		if(fd)
		{
			fwrite(m_Buffer,m_IFOSizekB*1024,1,fd);
			fclose(fd);
		}
	}
}

uchar* CIFOLoader::GetNumAudioStreams()
{
	return (m_Buffer+0x102);
}

AUDIO_ATTRIBUTES* CIFOLoader::GetAudioAttributes(uint ID)
{
	uchar *Ptr = m_Buffer+0x104+8*ID;
	return (AUDIO_ATTRIBUTES*) Ptr;
}

uchar* CIFOLoader::GoToCommands(ushort LU, ushort PGC)
{
	PGC_HEADER *Header = GetPGCHeader(LU, PGC);
	ushort Offset = FixEndian(Header->m_OffsetToCommands);
	if(Offset == 0)
		return 0;
	return (uchar*)Header + Offset;
}

ushort CIFOLoader::GetNumCommands(ushort LU, ushort PGC, int Type)
{
	uchar* Temp = GoToCommands(LU,PGC);
	if(Temp == 0)
		return 0;
	switch(Type)
	{
	case 0: return GetShort(Temp,0);
	case 1: return GetShort(Temp,2);
	case 2: return GetShort(Temp,4);
	case 3: return GetShort(Temp,0) + GetShort(Temp,2) + GetShort(Temp,4);
	}
	return 0;
}

uchar* CIFOLoader::GetCommands(ushort LU, ushort PGC, int Type, uchar** Commands)
{
	uchar* Ptr = GoToCommands(LU,PGC);
	if(Ptr == 0)
	{
		// 0 indicates there's no command whatsoever.
		if(Commands)
			*Commands = 0;
		return 0;
	}
	if(Commands)
	{
		switch(Type)
		{
		case 0: *Commands = Ptr + 8; break;
		case 1: *Commands = Ptr + 8 + 8*GetShort(Ptr,0); break; 
		case 2: *Commands = Ptr + 8 + 8*(GetShort(Ptr,0) + GetShort(Ptr,2)); break;
		}
	}
	return Ptr + Type*2;
}	

int CIFOLoader::AddCommand(ushort LU, ushort PGC, uchar* Command, int Type)
{
	ushort TotNumCommands = GetNumCommands(LU,PGC,3);
	if(TotNumCommands >= 128)
		return -1;

	// Change size of VTSM_PGCI_UT table. This will move other tables down if needed.
	uchar *Ptr;
	// This holds by how much we're moving things below us. It's normally 8 bytes.
	int Offset = 8;
	if(TotNumCommands == 0)
	{
		// No commands! There's no space for the numbers of pre/post/cell commands and the end address
		// Move things down by 8 bytes for the command table.
		PGC_HEADER *Header = GetPGCHeader(LU, PGC);
		MoveDataInTable(VTSM_PGCI_UT,(uchar*)Header+0xec,8);
		// Set the offset to the command tables.
		Header->m_OffsetToCommands = FixEndian((ushort)0xec);
		// Zero-out the number of pre, post and cell commands, and set the end byte to 0x7.
		PutShort((uchar*)Header,0xec,0);
		PutShort((uchar*)Header,0xec+2,0);
		PutShort((uchar*)Header,0xec+4,0);
		PutShort((uchar*)Header,0xec+6,0x07);
		// Now, every other PCG and LU tables below us will have to be moved by 8 more bytes.
		Offset += 8;
	}

	// Go there
	uchar *TableStart = FindTable(VTSM_PGCI_UT);
	uchar* NumCommands = GetCommands(LU, PGC, Type, &Ptr);

	// Move everything down below the command by 8 bytes. This will resize the table, and move others if needed.
	Ptr += 8*GetShort(NumCommands,0);
	MoveDataInTable(VTSM_PGCI_UT,Ptr,8);

	// Add the command...
	memcpy(Ptr,Command,8);

	// Now increase the number of commands and update the end address (offset 0x6 relative to
	// the number of precommands).
	AddShort(NumCommands,0,1); 
	AddShort(GetCommands(LU, PGC, 0, 0),6,8);	// Increase by 8 bytes.

	// Now update the program map, cell playback and cell position offsets.
	PGC_HEADER *Header = GetPGCHeader(LU, PGC);
	AddShort((uchar*)&Header->m_OffsetToPrograms,0,Offset);
	AddShort((uchar*)&Header->m_OffsetToCellPB,0,Offset);
	AddShort((uchar*)&Header->m_OffsetToCellPos,0,Offset);

	// Now update the PGC offsets defined in VTSM_LU for the PGCs after us.
	ushort NumPGC = GetNumPGC(LU);
	Ptr = GoToLU(LU);
	AddInt(Ptr,4,Offset); // Update end address!
	ushort pgc;
	for(pgc=PGC;pgc<NumPGC;pgc++)
	{
		AddInt(Ptr,12+pgc*8,Offset);
	}

	// Update the LU offsets defined in VTSM_PGCI_UT for the LUs after us
	ushort NumLU = GetNumLU();
	Ptr = FindTable(VTSM_PGCI_UT);
	AddInt(Ptr,0x4,Offset); // Update end address!
	ushort lu;
	for(lu=LU;lu<NumLU;lu++)
	{
		AddInt(Ptr,12+lu*8,Offset);
	}

// Just a test to see if I can really move tables around!
//	MoveDataInTable(VTSM_PGCI_UT,FindTable(VTSM_PGCI_UT)+GetEndAddress(VTSM_PGCI_UT)+1,2*1024);
//	AddInt(FindTable(VTSM_PGCI_UT),0x04,-2*1024);
	return (int)GetShort(NumCommands,0);
}

void CIFOLoader::UpdateVTS_C_ADT(CCell* TheCell, bool AddToVTSM_C_ADT, bool Menu)
{
	int i;
	uint LastVOBUEndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	for(i=0;i<(int)NumCells;i++)
	{
		ThisItem = &Table->m_Items[i];
		if(FixEndian(ThisItem->m_VOBID) != TheCell->m_VOBID || ThisItem->m_CELLID != TheCell->m_CELLID)\
			continue;
		// Found the cell, let's update the start and end LBAs.
		ThisItem->m_StartLBA = FixEndian((uint)TheCell->m_StartLBA);
		ThisItem->m_EndLBA = FixEndian(LastVOBUEndLBA);
		break;
	}
	if(i==(int)NumCells && AddToVTSM_C_ADT)
	{
		// The cell wasn't found in the VTSM_C_ADT, let's add one entry for it.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,sizeof(VTSM_C_ADTItem));
		ThisItem = &Table->m_Items[NumCells];
		ThisItem->m_VOBID = FixEndian(TheCell->m_VOBID);
		ThisItem->m_CELLID = TheCell->m_CELLID;
		ThisItem->m_StartLBA = FixEndian((uint)TheCell->m_StartLBA);
		ThisItem->m_EndLBA = FixEndian(LastVOBUEndLBA);
		// Change the last byte.
		AddInt((uchar*)&Table->m_LastByte,0,sizeof(VTSM_C_ADTItem));
	}
}

void CIFOLoader::RemoveFromVTS_C_ADT(CCell* TheCell, bool Menu)
{
	int i,j;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	for(i=j=0;i<(int)NumCells;i++)
	{
		Table->m_Items[j] = Table->m_Items[i];
		ThisItem = &Table->m_Items[i];
		if(FixEndian(ThisItem->m_VOBID) != TheCell->m_VOBID || ThisItem->m_CELLID != TheCell->m_CELLID)\
			j++;
	}
	if(j!=i)
	{
		// Change the table size.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,-(int)sizeof(VTSM_C_ADTItem));
		AddInt((uchar*)&Table->m_LastByte,0,-(int)sizeof(VTSM_C_ADTItem));
	}
	ushort Count=0;
	for(i=0;i<NumCells;i++)
	{
		// Go through list, and if you don't see this VOBID in the preceding cells, increment count.
		for(j=i-1;j>=0;j--)
		{
			if(Table->m_Items[j].m_VOBID == Table->m_Items[i].m_VOBID)
				break;
		}
		if(j==-1) Count++;
	}
	// Write that back into the table.
	Table->m_NumVOBID = FixEndian(Count);
}

void CIFOLoader::CleanupVTS_C_ADT(bool Menu)
{
	int i,j;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
	VTSM_C_ADTItem* ThisItem;
	ushort LUList[1024];
	ushort PGCList[1024];
	ushort CELLList[1024];
	for(i=j=0;i<(int)NumCells;i++)
	{
		// Go through list of cells, see if it's referenced in any of the PGCs...
		Table->m_Items[j] = Table->m_Items[i];
		ThisItem = &Table->m_Items[i];
		if(FindThisCell(FixEndian(ThisItem->m_VOBID),ThisItem->m_CELLID,
			(Menu)?LUList:0,PGCList,CELLList) != 0)
			j++;
	}
	if(j!=i)
	{
		// Change the table size.
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,(j-i)*(int)sizeof(VTSM_C_ADTItem));
		AddInt((uchar*)&Table->m_LastByte,0,(j-i)*sizeof(VTSM_C_ADTItem));
	}
	ushort Count=0;
	for(i=0;i<NumCells;i++)
	{
		// Go through list, and if you don't see this VOBID in the preceding cells, increment count.
		for(j=i-1;j>=0;j--)
		{
			if(Table->m_Items[j].m_VOBID == Table->m_Items[i].m_VOBID)
				break;
		}
		if(j==-1) Count++;
	}
	// Write that back into the table.
	Table->m_NumVOBID = FixEndian(Count);

}

void CIFOLoader::RebuildVTS_C_ADT(VTSM_C_ADTItem *Items, int NumCells, bool Menu)
{
	int i;

	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) FindTable(Menu?VTSM_C_ADT:VTS_C_ADT);
	if(Table == 0)
		return;
	ushort OrigNumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);

	// Clean up original table.
	memset(&Table->m_Items[0],0,OrigNumCells*sizeof(VTSM_C_ADTItem));

	// Then grow or shrink it!
	if(OrigNumCells != NumCells)
	{
		ChangeTableSize(Menu?VTSM_C_ADT:VTS_C_ADT,(NumCells-OrigNumCells)*(int)sizeof(VTSM_C_ADTItem));
		PutInt((uchar*)&Table->m_LastByte,0,NumCells*sizeof(VTSM_C_ADTItem)+8-1);
	}

	// This needs to be fixed for ILVU VTS because the cells have to be sorted by VOBID...
	// Then copy info into table...
	int j;
	ushort Count=0;
	int OutIndex = 0;
	// In this version, the cells are stored in VOBID/CID order, de-interleaved as they would if there
	// were ILVUs.
	for(i=0;i<NumCells;i++)
	{
		// Find if we already delt with this VOBID.
		for(j=i-1;j>=0;j--)
		{
			if(Items[j].m_VOBID == Items[i].m_VOBID)
				break;
		}
		if(j==-1)
		{
			// No, this is a new VOBID. Increment VOBID count, then write cell.
			Count++;
			Table->m_Items[OutIndex].m_VOBID = FixEndian(Items[i].m_VOBID);
			Table->m_Items[OutIndex].m_CELLID = Items[i].m_CELLID;
			Table->m_Items[OutIndex].m_StartLBA = FixEndian(Items[i].m_StartLBA);
			Table->m_Items[OutIndex++].m_EndLBA = FixEndian(Items[i].m_EndLBA);
			// Now look until the end of the input array and write all the cells have have the same
			// VOBID.
			for(j=i+1;j<NumCells;j++)
			{
				if(Items[j].m_VOBID == Items[i].m_VOBID)
				{
					// Same VOBID. Write it out....
					Table->m_Items[OutIndex].m_VOBID = FixEndian(Items[j].m_VOBID);
					Table->m_Items[OutIndex].m_CELLID = Items[j].m_CELLID;
					Table->m_Items[OutIndex].m_StartLBA = FixEndian(Items[j].m_StartLBA);
					Table->m_Items[OutIndex++].m_EndLBA = FixEndian(Items[j].m_EndLBA);
				}
			}
		}
	}
	// In the end, we should have written exactly how many we had on input.
	assert(OutIndex == NumCells);

	// Write the number of distinct VOBID back into the table.
	Table->m_NumVOBID = FixEndian(Count);
	return;
}


void CIFOLoader::UpdateCellInfo(CCell* TheCell, bool AddToVTSM_C_ADT, bool Menu)
{
	uint FirstVOBUStartLBA = TheCell->m_StartLBA;
	int LastNav = TheCell->FindPrevNavPack(TheCell->m_NumPacks);
	uint LastVOBUStartLBA = LastNav+TheCell->m_StartLBA;
	uint LastVOBUEndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;

	// Now, find every occurence of this cell.
	ushort LU[512], PGC[512], CELL[512];
	int NumFound = FindThisCell(TheCell->m_VOBID, TheCell->m_CELLID, (Menu)?LU:0, PGC, CELL);

	int i;
	for(i=0;i<NumFound;i++)
	{
		// Correct sector pointers for this PGC Cell.
		CELL_PLAYBACK_INFO* PBInfo = GetCellPlaybackInfo((Menu)?LU[i]:0,PGC[i],CELL[i]);
		PutInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0,FirstVOBUStartLBA);
		PutInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0,LastVOBUStartLBA);
		PutInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0,LastVOBUEndLBA);
		// Update the cell duration.
		CDVDTime CellDur(TheCell->m_EndPTS - TheCell->m_StartPTS);
		CellDur.WriteToBuffer((uchar*)&PBInfo->m_PBTime);
		// And the PGC time.
		UpdatePGCPlayBackTime((Menu)?LU[i]:0,PGC[i]);
	}

	// Now, look for the cell in the VTSM_C_ADT table..
	UpdateVTS_C_ADT(TheCell, AddToVTSM_C_ADT, Menu);

	// Add the cell's VOBUs to the list in the IFO.
	int NextNavPack = 0;
	while(NextNavPack != -1)
	{
		m_VOBUStartLBA[m_NumVOBUs] = TheCell->m_StartLBA + NextNavPack;
		m_NumVOBUs++;
		assert(MAX_NUM_VOBUS >= m_NumVOBUs);
		NextNavPack = TheCell->FindNextNavPack(NextNavPack);
	}
}

int FindMatchingNavPack(int ToMatch, int *InNavPacks, int *OutNavPacks, int NumNavPacks)
{
	int i;
	for(i=0;i<NumNavPacks;i++)
	{
		if(InNavPacks[i] >= ToMatch)
		{
//			assert(InNavPacks[i] == ToMatch);
			return OutNavPacks[i];
		}
	}
	return OutNavPacks[NumNavPacks-1];
}

void CIFOLoader::UpdateNavPackInfo(int *InNavPacks, int *OutNavPacks, int NumNavPacks, bool Menu)
{
	ushort LU = 0;
	ushort NumLUs = GetNumLU();
	if(!Menu) NumLUs = 1;

	for(LU=1;LU<=NumLUs;LU++)
	{
		ushort NumPGCs = GetNumPGC(Menu?LU:0);
		ushort PGC;
		for(PGC=1;PGC<=NumPGCs;PGC++)
		{
			ushort NumCells = GetNumCells(Menu?LU:0,PGC);
			ushort Cell;
			for(Cell=1;Cell<=NumCells;Cell++)
			{
				CELL_PLAYBACK_INFO* PBInfo = GetCellPlaybackInfo(Menu?LU:0,PGC,Cell);
				uint Temp = GetInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0);
				// Remember that the last navpack is a dummy one pointing after the end
				// of the VOB. We can't use that as a start LBA.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks-1);
				PutInt((uchar*)&PBInfo->m_FirstVOBUStartLBA,0,Temp);
				uint FirstVobuStart = Temp;

				Temp = GetInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0);
				// Same here.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks-1);
				uint LastVobuStart = Temp;
				assert(LastVobuStart>=FirstVobuStart);
				PutInt((uchar*)&PBInfo->m_LastVOBUStartLBA,0,Temp);

				Temp = GetInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0)+1;
				// But here, we can use the last navpack because we're looking for an end
				// LBA, not a start one.
				Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks);
				if(Temp >= 1) Temp --;
				assert(Temp>=LastVobuStart);
				PutInt((uchar*)&PBInfo->m_LastVOBUEndLBA,0,Temp);

				Temp = GetInt((uchar*)&PBInfo->m_FirstILVUEndLBA,0)+1;
				if(Temp > 1)
				{
					// same here...
					Temp = FindMatchingNavPack(Temp,InNavPacks,OutNavPacks,NumNavPacks);
					if(Temp >= 1) Temp --;
					PutInt((uchar*)&PBInfo->m_FirstILVUEndLBA,0,Temp);
				}
			}
		}
	}

	int i;
	m_NumVOBUs = 0;
	assert(MAX_NUM_VOBUS >= NumNavPacks-1);
	for(i=0;i<NumNavPacks;i++)
	{
		// Don't write twice the same guy!
		if(i == 0 || OutNavPacks[i] != OutNavPacks[i-1])
			m_VOBUStartLBA[m_NumVOBUs++] = OutNavPacks[i];
	}
	// NOTE I'm getting rid of the last one, which by convention is supposed to point past
	// the end of the VOB, as if there was a navpack there.
	if(m_NumVOBUs > 0)
		m_NumVOBUs--;
	
	// Now update the VOBU_ADMAP table.
	UpdateVOBUADMAP(Menu);
}

void CIFOLoader::UpdatePGCPlayBackTime(ushort LU, ushort PGC, bool FixZeroTimes)
{
	// Find this PGC...
	PGC_HEADER*	Header = GetPGCHeader(LU, PGC);

	// Get number of cells in this PGC
	uchar NumCells = GetNumCells(LU, PGC);

	// This will hold the total playback time of the PGC.
	CDVDTime TotalPGCTime(0);
	// Now go through all the cells in this PGC and add their playback times.
	uchar Cell;
	for(Cell=0;Cell<NumCells;Cell++)
	{
		CELL_PLAYBACK_INFO*	PBInfo = GetCellPlaybackInfo(LU, PGC, Cell+1);
		CDVDTime CellPBTime;
		// Read time, and add.
		CellPBTime.ReadFromBuffer((uchar*)(&PBInfo->m_PBTime));
		if(FixZeroTimes && CellPBTime.IsZero())
		{
			CellPBTime.m_frames = (VideoFormat == FORMAT_PAL) ? 12 : 18;
			CellPBTime.WriteToBuffer((uchar*)(&PBInfo->m_PBTime));
		}
		// Adjust elapsed time! But only if there's no angle, or this is the first block...
		if((PBInfo->m_Flags & 0x10) == 0 || 
			((PBInfo->m_Flags & 0x40) != 0 && (PBInfo->m_Flags & 0x80) == 0))
			TotalPGCTime += CellPBTime;
	}
	// Done, let's write the time back to the PGC info...
	TotalPGCTime.WriteToBuffer((uchar*)&Header->m_PlaybackTime);
}

// Renumbers all the VCIDs in all the pgc cells according to the map.
void CIFOLoader::RenumberVCID(VCIDMapItem* Map, int MapLength, bool Menu)
{
	int i,j,k,n;
	int NumLU = Menu?GetNumLU():1;
	// Go through all LUs.
	for(j=1;j<=NumLU;j++)
	{
		int LU = (Menu)?j:0;
		int NumPGC = GetNumPGC(LU);
		// And all PGCs.
		for(i=1;i<=NumPGC;i++)
		{
			int NumCell = GetNumCells(LU,i);
			// And all cells.
			for(k=1;k<=NumCell;k++)
			{
				// Now see if this VCID needs to be remapped.
				CELL_POS_INFO *Temp = GetCellPosInfo(LU,i,k);
				ushort VOBID = FixEndian(Temp->m_VOBID);
				uchar CELLID = Temp->m_CELLID;
				for(n=0;n<MapLength;n++)
				{
					if(Map[n].InVOBID == VOBID && Map[n].InCELLID == CELLID)
					{
						// Yes, remap it!
						Temp->m_VOBID = FixEndian(Map[n].OutVOBID);
						Temp->m_CELLID = Map[n].OutCELLID;
						break;
					}
				}
			}
		}
	}
}

void CIFOLoader::UpdateVOBUADMAP(bool Menu)
{
	uchar* Ptr;
	IFOTable VOBU_ADMAP = (Menu) ? VTSM_VOBU_ADMAP : VTS_VOBU_ADMAP;
	
	Ptr = FindTable(VOBU_ADMAP);
	if(Ptr == 0) return;
	
	// Find how many VOBUs were declared in the table.
	int NumVOBUS = (GetInt(Ptr,0) + 1)/4 - 1;
	
	// Now grow it, or shrink it, as necessary.
	int Delta = m_NumVOBUs - NumVOBUS;
	// If we shrink the table, replace the old values by 0s.
	int i;
	for(i=m_NumVOBUs;i<NumVOBUS;i++)
		PutInt(Ptr,(i+1)*4,0);
	// Then change the table size...
	ChangeTableSize(VOBU_ADMAP,Delta*4);

	// And write the VOBU info:
	for(i=0;i<m_NumVOBUs;i++)
		PutInt(Ptr,(i+1)*4,m_VOBUStartLBA[i]);

	// Finally, update the end byte of the table.
	PutInt(Ptr,0,3+m_NumVOBUs*4);
}

void CIFOLoader::UpdateTMAPTI(int *InNavPacks, int *OutNavPacks, int NumNavPacks)
{
	uchar* TMAPTI;
	TMAPTI = FindTable(VTS_TMAPTI);
	if(TMAPTI == 0) return;
	uchar* LastPtr = TMAPTI;

	ushort NumMaps = GetShort(TMAPTI,0);
	ushort i;

	for(i=0;i<NumMaps;i++)
	{
		// Find offset to map:
		uchar* Ptr = TMAPTI + GetInt(TMAPTI,8+4*i);
		ushort NumEntries = GetShort(Ptr,2);
		ushort j,k;
		uint LastEntry = 0xFFFFFFFF;
		for(j=k=0;j<NumEntries;j++)
		{
			// Get original entry, remember discontinuity, 
			uint OrigEntry = GetInt(Ptr,4+j*4);
			uint Discont = OrigEntry & 0x80000000;
			OrigEntry &= 0x7FFFFFFF;
			// Find matching output lba.
			uint NewEntry = FindMatchingNavPack(OrigEntry,InNavPacks,OutNavPacks,NumNavPacks-1);
			if(NewEntry != LastEntry)
			{
				// Put discontinuity back.
				LastEntry = NewEntry;
				NewEntry |= Discont;
				PutInt(Ptr,4+k*4,NewEntry);
				k++;
			}
		}
		// Done... We might have fewer entries! Write that back.
		PutShort(Ptr,2,k);
		// This is the end of this table: Ptr+4+4*k+3. Remember the furthest...
		LastPtr = max(LastPtr,Ptr+4+4*k+3);
	}
	// Now put back the size of the table, and change the size...
	uint OrigSize = GetInt(TMAPTI,4);
	uint NewSize = LastPtr-TMAPTI;
	ChangeTableSize(VTS_TMAPTI,(int)NewSize-(int)OrigSize);
	PutInt(TMAPTI,4,NewSize);
}

typedef struct
{
	// All sizes in sectors.
	int		m_VTSMenuFinalSize;
	int		m_IFOFinalSize;
	int		m_VTSTitleSize;
} VTSInfo;

void CIFOLoader::UpdateVTSSectors()
{
	if(!m_IsManager)
	{
		// Now update the size of the IFO table. To do this, look for all the tables and compute size
		// in LBA.
		uint MaxEndLBA = 0;
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_PTT_SRPT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_PGCI)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_PGCI_UT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_TMAPTI)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_VOBU_ADMAP)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_VOBU_ADMAP)/SECTOR_SIZE);
		// Now put that in 0x1c
		PutInt(m_Buffer,0x1c,MaxEndLBA);
		m_IFOSizekB = (MaxEndLBA+1)*2;
		// Set C0, first sector of VTS_xx_0.VOB.
		PutInt(m_Buffer,0xc0,m_IFOSizekB/2);

		// Set C4, first sector of VTS_xx_1.VOB.
		// Find size of VTS_xx_0.VOB
		int MenuSize = GetFileSizekB(ChangeExt(m_IFOFileName,".VOB"))/2;
		// Then find cumulative size of title vobs.
		int i;
		int TitleSize = 0;
		for(i=1;i<10;i++)
		{
			char Temp[512];
			sprintf(Temp,"%sVTS_%02d_%d.VOB",GetDirPath(m_IFOFileName),m_VTSNum,i);
			TitleSize += GetFileSizekB(Temp)/2;
		}
		PutInt(m_Buffer,0xc4,m_IFOSizekB/2+MenuSize);
		PutInt(m_Buffer,0x0c,2*m_IFOSizekB/2+MenuSize+TitleSize-1);
	}
	else
	{
		// Now update the size of the IFO table. To do this, look for all the tables and compute size
		// in LBA.
		uint MaxEndLBA = 0;
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(TT_SRPT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMGM_PGCI_UT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTS_PGCI)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_PGCI_UT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VMG_TXTDT_MG)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_C_ADT)/SECTOR_SIZE);
		MaxEndLBA = max(MaxEndLBA, GetAbsoluteEndAddress(VTSM_VOBU_ADMAP)/SECTOR_SIZE);
		// Now put that in 0x1c
		PutInt(m_Buffer,0x1c,(m_IFOSizekB/2)-1);
		// Set C0, first sector of VTS_xx_0.VOB.
		PutInt(m_Buffer,0xc0,m_IFOSizekB/2);

		int MenuSize = GetFileSizekB(ChangeExt(m_IFOFileName,".VOB"))/2;
		PutInt(m_Buffer,0x0c,2*m_IFOSizekB/2+MenuSize-1);
		// Now update the rest of the pointers in the TT_SRPT. For this I need to gather all the
		// file sizes. Start with the manager (this one!).
		VTSInfo* VTSInfos = new VTSInfo[100];
		VTSInfos[0].m_IFOFinalSize = m_IFOSizekB/2;
		VTSInfos[0].m_VTSMenuFinalSize = MenuSize;
		VTSInfos[0].m_VTSTitleSize = 0;
		uint VMGSize = 2*VTSInfos[0].m_IFOFinalSize+VTSInfos[0].m_VTSMenuFinalSize;

		// Then the vobs.
		uint StartOfVTS[100];
		StartOfVTS[0] = 0;
		int i,j;
		for(i=1;i<100;i++)
		{
			VTSInfos[i].m_IFOFinalSize = GetFileSizekB(CatStr(GetDirPath(m_IFOFileName),IFONames[i]))/2;
			VTSInfos[i].m_VTSMenuFinalSize = GetFileSizekB(ChangeExt(CatStr(GetDirPath(m_IFOFileName),IFONames[i]),".VOB"))/2;
			int Size = 0;
			for(j=1;j<10;j++)
			{
				char Temp[512];
				sprintf(Temp,"%sVTS_%02d_%d.VOB",GetDirPath(m_IFOFileName),i,j);
				int ThisSize = GetFileSizekB(Temp)/2;
				if(ThisSize == 0 && j > 0)
					break;
				Size += ThisSize;
			}
			VTSInfos[i].m_VTSTitleSize = Size;
			StartOfVTS[i] = StartOfVTS[i-1]+2*VTSInfos[i-1].m_IFOFinalSize+VTSInfos[i-1].m_VTSMenuFinalSize+
				VTSInfos[i-1].m_VTSTitleSize;
		}

		uchar* Ptr = FindTable(TT_SRPT);
		if(Ptr)
		{
			uint NumTitles = GetShort(Ptr,0);
			for(i=0;i<(int)NumTitles;i++)
			{
				// Get VTS number:
				uchar VTSNumber = Ptr[8+i*12+6];
				PutInt(Ptr,8+i*12+8,StartOfVTS[VTSNumber]);
			}
		}
		delete[] VTSInfos;
	}

}