/*
 *  MenuShrink and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  MenuShrink and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "stdafx.h"
#include "MenuShrink.h"
#include "MenuShrinkDlg.h"
#include "MiscDialogs.h"
#include "FrameSelectDialog.h"
#include "MenuShrinkEngine.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CFrameSelectDialog*  FrameSelectDlg = NULL;

bool	WindowsMode;	// Indicates we're on a windows machine (not wine).


CRichEditCtrl*	TheLogEdit = NULL;
bool	QuitNow = 0;
int		ManualSelect = 0;
int		NoOverlayMode = 0;
// Pick language according to defines.
#ifdef _MYENGLISH
int		CurrentLangage = ENGLISH;
#else
int		CurrentLangage = FRENCH;
#endif
char	MessageOut[512];

// Error == 1 is used to highlight the message in red...
void LogMessage(char* AMessage, int Langage, int Error)
{
	if(QuitNow || CurrentLangage != Langage)
		return;

	CHARFORMAT cf;
	cf.dwMask= CFM_COLOR | CFM_UNDERLINE;
	cf.dwEffects= 0;
	COLORREF Color = RGB(0,0,255);

	// This is the log function that replaces the default log function in the MenuShrinkEngine.
	int i,j;
	int L = strlen(AMessage);
	i=j=0;
	if(AMessage[0] == '\n') i++;
	for(;i<L;i++,j++)
	{
		if(WindowsMode && AMessage[i] == '\n')
		{
			MessageOut[j++] = '\r';
			MessageOut[j] = '\n';
			if(AMessage[i+1] == '\n')
				i++;
		}
		else
			MessageOut[j] = AMessage[i];
	}
	MessageOut[j] = '\0';

	// Change color for warnings, notes, errors etc.
	if(strncmp(AMessage,"NOTE", 4) == 0 || strncmp(AMessage,"WARNING",7) == 0)
		Color = RGB(0,134,37);
	if(Error == 1)
		// Error messages in red.
		Color = RGB(255,0,0);
	else if(strncmp(AMessage,"Shrinking", 5) == 0 || strncmp(AMessage,"Traitement", 5) == 0)
	{
		// Underline this line...
		cf.dwMask |= CFM_UNDERLINE;
		cf.dwEffects |= CFE_UNDERLINE;
	}

	int N = TheLogEdit->GetWindowTextLength();
	TheLogEdit->SetSel(-1,-1);
	TheLogEdit->ReplaceSel(MessageOut);
	cf.crTextColor=Color;
	// Select last line and apply color (or style).
	TheLogEdit->SetSel(N,TheLogEdit->GetWindowTextLength());
	TheLogEdit->SetSelectionCharFormat(cf);
	TheLogEdit->LineScroll(2);
}

void MessageFunc(char* AMessage, int Langage)
{
	LogMessage(AMessage, Langage, 0);
}

void ErrorFunc(char* AMessage, int Langage)
{
	LogMessage(AMessage, Langage, 1);
}

int UISelectFrame(int VobId, int CellId, int StartLBA, int EndLBA, int Default, int NumLBA, int NormallySkipped)
{
	if(ManualSelect == 0)
		return Default;
	char Temp[512];
	if(CurrentLangage == ENGLISH)
		sprintf(Temp,"Select I-frame for Cell %d/%d -- Size %.2f MB",VobId,CellId,(float)NumLBA/512);
	else
		sprintf(Temp,"Choisir l'image pour la cellule %d/%d -- Taille %.2f MB",VobId,CellId,(float)NumLBA/512);
	FrameSelectDlg->m_Title = Temp;
	FrameSelectDlg->m_FileName = ShrinkEngine->m_VTSCurrentlyProcessing;
	FrameSelectDlg->m_StartLBA = StartLBA;
	FrameSelectDlg->m_EndLBA = EndLBA;
	FrameSelectDlg->m_DefaultLBA = Default;
	FrameSelectDlg->m_VOBID = VobId;
	FrameSelectDlg->m_CELLID= CellId;
	FrameSelectDlg->m_NormallySkipped = NormallySkipped;
	if(FrameSelectDlg->m_VTSIndex != ShrinkEngine->m_CurrentVTSIndex)
		FrameSelectDlg->m_Auto = 0;
	FrameSelectDlg->m_VTSIndex = ShrinkEngine->m_CurrentVTSIndex;

	if(FrameSelectDlg->m_Auto)
		return FrameSelectDlg->GetSelectedLBA();;
	FrameSelectDlg->DoModal();
	return FrameSelectDlg->GetSelectedLBA();
}

/////////////////////////////////////////////////////////////////////////////
// CMenuShrinkDlg dialog


CMenuShrinkDlg::CMenuShrinkDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CMenuShrinkDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CMenuShrinkDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	m_OpenAtStartup = NULL;
	m_VTSToOpen = -1;
}

void CMenuShrinkDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CMenuShrinkDlg)
	DDX_Control(pDX, IDC_KEEPAUDIO, m_KeepAudio);
	DDX_Control(pDX, IDC_MESSAGE, m_Message);
	DDX_Control(pDX, IDC_PROGRESS, m_Progress);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CMenuShrinkDlg, CDialog)
	//{{AFX_MSG_MAP(CMenuShrinkDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_PROCESS, OnProcess)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_RESTORE, OnRestore)
	ON_BN_CLICKED(IDC_SHRINK, OnShrink)
	ON_BN_CLICKED(IDC_SHOWLOW, OnShowlow)
	ON_BN_CLICKED(IDC_KEEPAUDIO, OnKeepaudio)
	ON_BN_CLICKED(IDC_SHOWADVANCED, OnShowadvanced)
	ON_BN_CLICKED(IDC_POWERDVD, OnPowerdvd)
	ON_BN_CLICKED(IDC_SHOWABOUT, OnShowabout)
	ON_WM_CTLCOLOR()
	ON_BN_CLICKED(IDC_SHOWPATH, OnShowpath)
	ON_BN_CLICKED(IDC_OPENFILES, OnOpenfiles)
	ON_BN_CLICKED(IDC_OPEN, OnOpen)
	//}}AFX_MSG_MAP
	ON_WM_DROPFILES()
	ON_NOTIFY_EX( TTN_NEEDTEXT, 0, OnToolTipNotify )
	ON_COMMAND_RANGE(IDC_F1, IDC_F5, OnFKey)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMenuShrinkDlg message handlers

BOOL CMenuShrinkDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here
	DragAcceptFiles();

	// Create directory for Ini file, if it isn't there yet!
	WindowsMode = true;
	if(getenv("APPDATA") != 0)
	{
		sprintf(m_IniFile,"%s/MenuShrink",getenv("APPDATA"));
		mkdir(m_IniFile);
		sprintf(m_IniFile,"%s/MenuShrink/MenuShrink.ini",getenv("APPDATA"));
	}
	else
	{
		char Temp[1024];
		GetCurrentDirectory(1024,Temp);
		sprintf(m_IniFile,"%s/.MenuShrink",Temp);
		mkdir(m_IniFile);
		sprintf(m_IniFile,"%s/.MenuShrink/MenuShrink.ini",Temp);
		WindowsMode = false;
	}

	// By setting this function pointer to our own routine, we make sure messages are received here,
	// instead of send to the default printf() function of MenuShrinkEngine.
	pErrorFunction = ErrorFunc;
	pMessageFunction = MessageFunc;
	m_Running = 0;
	m_ReturnedError = 0;

	EnableToolTips(true);

	strcpy(MessageOut,"");

	m_Timer = SetTimer(0,200,NULL);
	SetWindowText(VERSION);

	m_LogDialog = new CLogDialog();
	m_LogDialog->CDialog::Create(IDD_LOGDIALOG);

	m_AdvancedDialog = new CAdvancedDialog();
	m_AdvancedDialog->CDialog::Create(IDD_ADVANCED);
	m_AdvancedDialog->EnableToolTips(true);
	if(CurrentLangage != ENGLISH)
		m_AdvancedDialog->SetWindowText("Options avances");

	m_PathDialog = new CPathDialog();
	m_PathDialog->CDialog::Create(IDD_PATHDIALOG);

	FrameSelectDlg = new CFrameSelectDialog();
	FrameSelectDlg->EnableToolTips(true);

	// Now for default value stuff.
	char TempC[256];
	int Temp = GetPrivateProfileInt("MenuShrink","LogOpen",0,m_IniFile);
	m_LogDialog->ShowWindow(Temp?SW_SHOW:SW_HIDE); 
	Temp = GetPrivateProfileInt("MenuShrink","AdvancedOpen",0,m_IniFile);
	m_AdvancedDialog->ShowWindow(Temp?SW_SHOW:SW_HIDE); 

	int x,y,cx,cy;
	GetPrivateProfileString("MenuShrink","LogPos","46 114 649 287",TempC,512,m_IniFile);
	sscanf(TempC,"%d %d %d %d",&x,&y,&cx,&cy);
	m_LogDialog->SetWindowPos(NULL,x,y,cx,cy,SWP_NOZORDER);

	GetPrivateProfileString("MenuShrink","AdvPos","696 114",TempC,512,m_IniFile);
	sscanf(TempC,"%d %d",&x,&y);
	m_AdvancedDialog->SetWindowPos(NULL,x,y,0,0,SWP_NOZORDER|SWP_NOSIZE);

	GetPrivateProfileString("MenuShrink","MainPos","380 400",TempC,512,m_IniFile);
	sscanf(TempC,"%d %d",&x,&y);
	SetWindowPos(NULL,x,y,0,0,SWP_NOZORDER|SWP_NOSIZE);

	GetPrivateProfileString("MenuShrink","FrameSelectPos","386 401 828 781",TempC,512,m_IniFile);
	sscanf(TempC,"%d %d %d %d",&x,&y,&cx,&cy);
	FrameSelectDlg->m_MyRect = CRect(x,y,cx,cy);

	GetPrivateProfileString("MenuShrink","KeepAudio","0",TempC,512,m_IniFile);
	ShrinkEngine->m_KeepAudio = (strcmp(TempC,"1") == 0);
	((CButton*)GetDlgItem(IDC_KEEPAUDIO))->SetCheck(ShrinkEngine->m_KeepAudio);

	int frame = GetPrivateProfileInt("MenuShrink","FrameSelect",85,m_IniFile);
	ShrinkEngine->m_FrameSelect = (FrameSelect) frame;

	ManualSelect = GetPrivateProfileInt("MenuShrink","ManualSelect",0,m_IniFile);
	m_AdvancedDialog->m_WarnAboutPreview = GetPrivateProfileInt("MenuShrink","WarnPreview",1,m_IniFile);

	// This overrides the frame setting if there's no version number. A bit kludgy.
	GetPrivateProfileString("MenuShrink","Version","",TempC,512,m_IniFile);
	bool NewVersion = (strcmp(TempC,VERSION) != 0);
	if(NewVersion)
		ShrinkEngine->m_FrameSelect = 85;

	GetPrivateProfileString("MenuShrink","Shrink","C:\\Program Files\\DVD Shrink\\DVD Shrink 3.2.exe",
		m_PathDialog->m_PathToShrink,MAX_PATH,m_IniFile);
	GetPrivateProfileString("MenuShrink","Preview","C:\\Program Files\\CyberLink\\PowerDVD\\PowerDVD.exe",
		m_PathDialog->m_PathToPowerDVD,512,m_IniFile);

	Temp = GetPrivateProfileInt("MenuShrink","FixSubPic",1,m_IniFile);
	ShrinkEngine->m_FixSubPicStops = Temp==1;

	ShrinkEngine->m_MinButtonSize = GetPrivateProfileInt("MenuShrink","MinButtonSize",5,m_IniFile);

	Temp = GetPrivateProfileInt("MenuShrink","InfiniteStills",0,m_IniFile);
	ShrinkEngine->m_UseInfiniteStills = Temp==1;

	ShrinkEngine->m_SkipButtonlessCells = GetPrivateProfileInt("MenuShrink","SkipButtonlessClips",1,m_IniFile);

	ShrinkEngine->m_OneSecondStill = GetPrivateProfileInt("MenuShrink","OneSecondStill",0,m_IniFile);

	NoOverlayMode = GetPrivateProfileInt("MenuShrink","NoOverlayMode",0,m_IniFile);
	if(!WindowsMode)
		NoOverlayMode = 1;

	// Paths for tools...
	GetPrivateProfileString("MenuShrink","F1","",m_PathDialog->m_Tool1,MAX_PATH,m_IniFile);
	GetPrivateProfileString("MenuShrink","F2","",m_PathDialog->m_Tool2,MAX_PATH,m_IniFile);
	GetPrivateProfileString("MenuShrink","F3","",m_PathDialog->m_Tool3,MAX_PATH,m_IniFile);
	GetPrivateProfileString("MenuShrink","F4","",m_PathDialog->m_Tool4,MAX_PATH,m_IniFile);
	GetPrivateProfileString("MenuShrink","F5","",m_PathDialog->m_Tool5,MAX_PATH,m_IniFile);

	m_AdvancedDialog->RefreshUI();
	m_PathDialog->RefreshUI();

	TheLogEdit = (CRichEditCtrl*) m_LogDialog->GetDlgItem(IDC_EDITLOG);
//	TheLogEdit->SetLimitText(1000000);
	InitializeCriticalSection(&m_csLock);

	FrameSelectFunction = UISelectFrame;

	if(m_OpenAtStartup != NULL)
	{
		OpenDVD(m_OpenAtStartup);
		if(m_VTSToOpen != -1)
		{
			// Let's mark only the specified VTS for processing.
			int i;
			for(i=0;i<ShrinkEngine->m_LastVTS+1;i++)
			{
				ShrinkEngine->m_VTSInfo[i].m_SkipThisVTS = 1;
			}
			ShrinkEngine->m_VTSInfo[m_VTSToOpen].m_SkipThisVTS = 0;
			m_AdvancedDialog->UpdateSkips();
			m_VTSToOpen = -1;
		}
	}

	AdjustToLanguage();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

// 

void CMenuShrinkDlg::AdjustToLanguage()
{
	if(CurrentLangage == ENGLISH)
		return;
	// Let's change all the strings
	if(CurrentLangage == FRENCH)
	{
		SetDlgItemText(IDC_OPEN,"Ouvrir");
		SetDlgItemText(IDC_PROCESS,"Dmarrer");
		SetDlgItemText(IDC_RESTORE,"Retourner  l'original");
		SetDlgItemText(IDC_SHOWADVANCED,"Options avances");
		SetDlgItemText(IDC_SHOWABOUT,"A propos/Aide");
		SetDlgItemText(IDC_KEEPAUDIO,"Garder l'audio");
		SetDlgItemText(IDC_TOOLFRAME,"Outils");
		SetDlgItemText(IDC_SHOWPATH,"Choix des chemins");
		SetDlgItemText(IDC_DRAGDROP,"Glissez/dplacer un rpertoire, dmarrez, c'est tout!");
		SetDlgItemText(IDC_OPENFILES,"Ouvrir explorer");
	}
}


BOOL CMenuShrinkDlg::DestroyWindow() 
{
	QuitNow = 1;
	ShrinkEngine->m_AbortNow = 1;
	EnterCriticalSection(&m_csLock);
	char Temp[256];
	WritePrivateProfileString("MenuShrink","LogOpen",m_LogDialog->IsWindowVisible() ? "1" : "0",m_IniFile);
	WritePrivateProfileString("MenuShrink","AdvancedOpen",m_AdvancedDialog->IsWindowVisible() ? "1" : "0",m_IniFile);
	RECT Rect;
	m_LogDialog->GetWindowRect(&Rect);
	sprintf(Temp,"%d %d %d %d",Rect.left,Rect.top,Rect.right-Rect.left,Rect.bottom-Rect.top);
	WritePrivateProfileString("MenuShrink","LogPos",Temp,m_IniFile);

	Rect = FrameSelectDlg->m_MyRect;
	sprintf(Temp,"%d %d %d %d",Rect.left,Rect.top,Rect.right,Rect.bottom);
	WritePrivateProfileString("MenuShrink","FrameSelectPos",Temp,m_IniFile);

	WritePrivateProfileString("MenuShrink","Version",VERSION,m_IniFile);

	m_AdvancedDialog->GetWindowRect(&Rect);
	sprintf(Temp,"%d %d",Rect.left,Rect.top);
	WritePrivateProfileString("MenuShrink","AdvPos",Temp,m_IniFile);

	GetWindowRect(&Rect);
	sprintf(Temp,"%d %d",Rect.left,Rect.top);
	WritePrivateProfileString("MenuShrink","MainPos",Temp,m_IniFile);
	
	WritePrivateProfileString("MenuShrink","KeepAudio",ShrinkEngine->m_KeepAudio?"1":"0",m_IniFile);
	WritePrivateProfileString("MenuShrink","ManualSelect",ManualSelect?"1":"0",m_IniFile);
	sprintf(Temp,"%d",(int)(ShrinkEngine->m_FrameSelect));
	WritePrivateProfileString("MenuShrink","FrameSelect",Temp,m_IniFile);

	WritePrivateProfileString("MenuShrink","Shrink",m_PathDialog->m_PathToShrink,m_IniFile);
	WritePrivateProfileString("MenuShrink","Preview",m_PathDialog->m_PathToPowerDVD,m_IniFile);
	WritePrivateProfileString("MenuShrink","WarnPreview",m_AdvancedDialog->m_WarnAboutPreview ? "1":"0",m_IniFile);

	WritePrivateProfileString("MenuShrink","FixSubPic",ShrinkEngine->m_FixSubPicStops?"1":"0",m_IniFile);
	sprintf(Temp,"%d",(int)(ShrinkEngine->m_MinButtonSize));
	WritePrivateProfileString("MenuShrink","MinButtonSize",Temp,m_IniFile);

	WritePrivateProfileString("MenuShrink","InfiniteStills",ShrinkEngine->m_UseInfiniteStills?"1":"0",m_IniFile);
	WritePrivateProfileString("MenuShrink","SkipButtonlessClips",ShrinkEngine->m_SkipButtonlessCells?"1":"0",m_IniFile);

	WritePrivateProfileString("MenuShrink","F1",m_PathDialog->m_Tool1,m_IniFile);
	WritePrivateProfileString("MenuShrink","F2",m_PathDialog->m_Tool2,m_IniFile);
	WritePrivateProfileString("MenuShrink","F3",m_PathDialog->m_Tool3,m_IniFile);
	WritePrivateProfileString("MenuShrink","F4",m_PathDialog->m_Tool4,m_IniFile);
	WritePrivateProfileString("MenuShrink","F5",m_PathDialog->m_Tool5,m_IniFile);

	// Save I-frame selection info, if any.
	FrameSelectDlg->SaveData(ShrinkEngine->m_WorkingDirFullPath);

	m_LogDialog->DestroyWindow();
	delete m_LogDialog;
	m_AdvancedDialog->DestroyWindow();
	delete m_AdvancedDialog;
	m_PathDialog->DestroyWindow();
	delete m_PathDialog;
	FrameSelectDlg->DestroyWindow();
	delete FrameSelectDlg;
	LeaveCriticalSection(&m_csLock);
	DeleteCriticalSection(&m_csLock);
	delete ShrinkEngine;
	return CDialog::DestroyWindow();
}


void CMenuShrinkDlg::OnDropFiles(HDROP hDropInfo) 
{
	SetForegroundWindow();      // activate us first !
	POINT ThePoint;
	DragQueryPoint(hDropInfo,&ThePoint);

	UINT nFiles = ::DragQueryFile(hDropInfo, (UINT)-1, NULL, 0);
	TCHAR szFileName[_MAX_PATH];
	::DragQueryFile(hDropInfo, 0, szFileName, _MAX_PATH);
	OpenDVD(szFileName);
	::DragFinish(hDropInfo);
}

void CMenuShrinkDlg::OpenDVD(char* Path)
{
	FrameSelectDlg->SaveData(ShrinkEngine->m_WorkingDirFullPath);
	m_ReturnedError = 0;
	ResolveIt(m_hWnd, Path);
	TheLogEdit->SetWindowText("");
	ShrinkEngine->ScanDVDFolder(Path, true);
	CString Temp;
	if(CurrentLangage == ENGLISH)
	{
		Temp.Format("Scanned %s -- %d VTS found - Total %.2fMB",ShrinkEngine->m_WorkingDirFullPath,
			ShrinkEngine->m_LastVTS+1,2*(float)ShrinkEngine->m_NumSectorsToProcess/1024);
	}
	else
	{
		Temp.Format("Lecture de %s -- %d VTS trouvs - Total %.2fMB",ShrinkEngine->m_WorkingDirFullPath,
			ShrinkEngine->m_LastVTS+1,2*(float)ShrinkEngine->m_NumSectorsToProcess/1024);
	}
	SetWindowText(Temp);
	MessageFunc(Temp.GetBuffer(1000),0);
	MessageFunc(" \n",0);
	ShrinkEngine->GetOveralDVDSize();
	m_AdvancedDialog->UpdateSkips();
	if(strcmp(ShrinkEngine->m_WorkingDirFullPath,"") != 0)
		WritePrivateProfileString("MenuShrink","LastDVD",ShrinkEngine->m_WorkingDirFullPath,m_IniFile);
	FrameSelectDlg->LoadData(ShrinkEngine->m_WorkingDirFullPath);
	m_AdvancedDialog->EnableUI(true);
}

UINT ShrinkEngineProcessFunction( LPVOID pParam )
{
	CMenuShrinkDlg* Dialog = (CMenuShrinkDlg*) pParam;
	Dialog->m_Running = 1;
	Dialog->m_ReturnedError = 0;
	Dialog->KillTimer(Dialog->m_Timer);
	Dialog->m_Timer = Dialog->SetTimer(0,200,NULL);
	FrameSelectDlg->m_Auto = 0;
	EnterCriticalSection(&Dialog->m_csLock);
	Dialog->DragAcceptFiles(FALSE);

	int ret = ShrinkEngine->ProcessFullDVD();
	Dialog->m_ReturnedError = ret != 0;
	
	if(QuitNow == 0)
	{
		Dialog->m_LogDialog->SetWindowText("Log");
		ShrinkEngine->m_AbortNow = 0;
		Dialog->m_AdvancedDialog->EnableUI(true);
		Dialog->m_Running = 0;
		if(CurrentLangage == FRENCH)
			Dialog->SetDlgItemText(IDC_PROCESS,"Dmarrer");
		else
			Dialog->SetDlgItemText(IDC_PROCESS,"Process");
		Dialog->KillTimer(Dialog->m_Timer);
		Dialog->m_Timer = Dialog->SetTimer(0,200,NULL);
		ShrinkEngine->m_SectorsProcessedSoFar=0;
		Dialog->DragAcceptFiles(TRUE);
	}
	LeaveCriticalSection(&Dialog->m_csLock);
	return 0;
}

void CMenuShrinkDlg::OnProcess() 
{
	if(m_Running == 0)
	{
		TheLogEdit->SetWindowText("");
		if(CurrentLangage == FRENCH)
			SetDlgItemText(IDC_PROCESS,"Annuler");
		else
			SetDlgItemText(IDC_PROCESS,"Abort");
		m_AdvancedDialog->EnableUI(false);
		AfxBeginThread( ShrinkEngineProcessFunction, this);
	}
	else
		ShrinkEngine->m_AbortNow = 1;
}

void CMenuShrinkDlg::OnTimer(UINT nIDEvent) 
{
	m_Progress.SetPos(100*ShrinkEngine->m_SectorsProcessedSoFar / (1+ShrinkEngine->m_NumSectorsToProcess));
	int j;
	char ThisMessage[512];
	char Temp[512];
	strcpy(ThisMessage,MessageOut);
	for(j=strlen(ThisMessage)-1;j>=0;j--)
	{
		if(ThisMessage[j]=='\n' || ThisMessage[j]=='\r')
			ThisMessage[j]=' ';
	}
	if(m_Running && ShrinkEngine->m_VTSCurrentlyProcessing != NULL)
	{
		if(CurrentLangage == ENGLISH)
			sprintf(Temp,"Processing %s",ShrinkEngine->m_InFileName);
		else
			sprintf(Temp,"Traitement de %s",ShrinkEngine->m_InFileName);
		m_LogDialog->SetWindowText(Temp);
	}
	UpdateUI();
	m_Message.SetWindowText(ThisMessage);
	CDialog::OnTimer(nIDEvent);
}

void CMenuShrinkDlg::UpdateUI()
{
	// Disable other buttons if needed
	GetDlgItem(IDC_PROCESS)->EnableWindow(ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_SHRINK)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1 && 
		strcmp(m_PathDialog->m_PathToShrink,"") != 0);
	GetDlgItem(IDC_RESTORE)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_KEEPAUDIO)->EnableWindow(!m_Running);
	GetDlgItem(IDC_OPEN)->EnableWindow(!m_Running);
	GetDlgItem(IDC_POWERDVD)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_F1)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_F2)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_F3)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_F4)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_F5)->EnableWindow(!m_Running && ShrinkEngine->m_LastVTS != -1);
	GetDlgItem(IDC_OPENFILES)->EnableWindow(!m_Running && 
		strcmp(ShrinkEngine->m_WorkingDirFullPath,"") != 0);

	char Temp[512];
	_splitpath(m_PathDialog->m_PathToPowerDVD,NULL,NULL,Temp,NULL); SetDlgItemText(IDC_POWERDVD,Temp);
	GetDlgItem(IDC_POWERDVD)->ShowWindow(strcmp(m_PathDialog->m_PathToPowerDVD,"") ? SW_SHOW : SW_HIDE);
	GetDlgItem(IDC_SHRINK)->ShowWindow(strcmp(m_PathDialog->m_PathToShrink,"") ? SW_SHOW : SW_HIDE);

	bool ShowTool[6];
	ShowTool[1] = strcmp(m_PathDialog->m_Tool1N,"") != 0;
	ShowTool[2] = strcmp(m_PathDialog->m_Tool2N,"") != 0;
	ShowTool[3] = strcmp(m_PathDialog->m_Tool3N,"") != 0;
	ShowTool[4] = strcmp(m_PathDialog->m_Tool4N,"") != 0;
	ShowTool[5] = strcmp(m_PathDialog->m_Tool5N,"") != 0;

	_splitpath(m_PathDialog->m_Tool1,NULL,NULL,m_PathDialog->m_Tool1N,NULL); SetDlgItemText(IDC_F1,m_PathDialog->m_Tool1N);
	GetDlgItem(IDC_F1)->ShowWindow(ShowTool[1] ? SW_SHOW : SW_HIDE);
	_splitpath(m_PathDialog->m_Tool2,NULL,NULL,m_PathDialog->m_Tool2N,NULL); SetDlgItemText(IDC_F2,m_PathDialog->m_Tool2N);
	GetDlgItem(IDC_F2)->ShowWindow(ShowTool[2] ? SW_SHOW : SW_HIDE);
	_splitpath(m_PathDialog->m_Tool3,NULL,NULL,m_PathDialog->m_Tool3N,NULL); SetDlgItemText(IDC_F3,m_PathDialog->m_Tool3N);
	GetDlgItem(IDC_F3)->ShowWindow(ShowTool[3] ? SW_SHOW : SW_HIDE);
	_splitpath(m_PathDialog->m_Tool4,NULL,NULL,m_PathDialog->m_Tool4N,NULL); SetDlgItemText(IDC_F4,m_PathDialog->m_Tool4N);
	GetDlgItem(IDC_F4)->ShowWindow(ShowTool[4] ? SW_SHOW : SW_HIDE);
	_splitpath(m_PathDialog->m_Tool5,NULL,NULL,m_PathDialog->m_Tool5N,NULL); SetDlgItemText(IDC_F5,m_PathDialog->m_Tool5N);
	GetDlgItem(IDC_F5)->ShowWindow(ShowTool[5] ? SW_SHOW : SW_HIDE);
}

// _________________________ BUTTON STUFF _______________

void CMenuShrinkDlg::OnRestore() 
{
	m_ReturnedError = 0;
	ShrinkEngine->RestoreBackup();	
}
void CMenuShrinkDlg::OnShrink() 
{
	ShellExecute(NULL, "open", m_PathDialog->m_PathToShrink, 
		"VIDEO_TS.IFO", NULL, SW_SHOWNORMAL);
}
void CMenuShrinkDlg::OnShowlow() 
{
	m_LogDialog->ShowWindow(m_LogDialog->IsWindowVisible()?SW_HIDE:SW_SHOW);
}
void CMenuShrinkDlg::OnKeepaudio() 
{
	ShrinkEngine->m_KeepAudio = ((CButton*)GetDlgItem(IDC_KEEPAUDIO))->GetCheck();
	m_AdvancedDialog->RefreshUI();
}
void CMenuShrinkDlg::OnShowadvanced() 
{
	m_AdvancedDialog->ShowWindow(m_AdvancedDialog->IsWindowVisible()?SW_HIDE:SW_SHOW);
}
void CMenuShrinkDlg::OnPowerdvd() 
{
	char Temp[512];
	// Create path to VIDEO_TS.IFO!
	sprintf(Temp,"\"%s\\VIDEO_TS.IFO\"",ShrinkEngine->m_WorkingDirFullPath);
	if(strcmp(m_PathDialog->m_PathToPowerDVD,"") != 0)
		ShellExecute(NULL, "open", m_PathDialog->m_PathToPowerDVD, Temp, NULL, SW_SHOWNORMAL);
}
void CMenuShrinkDlg::OnShowabout() 
{
	CAboutDlg dlgAbout;
	dlgAbout.DoModal();
}
void CMenuShrinkDlg::OnShowpath() 
{
	m_PathDialog->ShowWindow(m_PathDialog->IsWindowVisible()?SW_HIDE:SW_SHOW);
}
void CMenuShrinkDlg::OnOpenfiles() 
{
	ShellExecute(NULL, "open", ShrinkEngine->m_WorkingDirFullPath, "", NULL, SW_SHOWNORMAL);
}

// _________________________ MFC CLUTTER _______________

void CMenuShrinkDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.
void CMenuShrinkDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CMenuShrinkDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

bool CMenuShrinkDlg::OnToolTipNotify( UINT id, NMHDR * pNMHDR, LRESULT * pResult )
{
	// Multiline tooltips: http://www.thecodeproject.com/miscctrl/tooltipex.asp
    TOOLTIPTEXT *pTTT = (TOOLTIPTEXT *)pNMHDR;
    UINT nID =pNMHDR->idFrom;
	char Temp[512];
    if (pTTT->uFlags & TTF_IDISHWND)
    {
        // idFrom is actually the HWND of the tool
        nID = ::GetDlgCtrlID((HWND)nID);
		if(CurrentLangage == ENGLISH)
		{
			switch(nID)
			{
			case IDC_SHRINK : pTTT->lpszText = "Launch DVD Shrink - Set path in path dialog"; break; 
			case IDC_POWERDVD : pTTT->lpszText = "Launch soft player - Set path in path dialog. Works with media player classic, PowerDVD 6.0, probably more..."; break; 
			case IDC_RESTORE : pTTT->lpszText = "Undo all processing and restore backup files"; break; 
			case IDC_SHOWADVANCED : pTTT->lpszText = "Show panel with advanced options"; break; 
			case IDC_SHOWLOW : pTTT->lpszText = "Show log panel with all MenuShrink messages"; break; 
			case IDC_SHOWPATH : pTTT->lpszText = "Show path dialog, set tool paths etc"; break; 
			case IDC_PROCESS : pTTT->lpszText = "Start shrinking the menus"; break; 
			case IDC_KEEPAUDIO : pTTT->lpszText = "Check to create still menus with audio. Uncheck for slightly smaller silent menus"; break; 
			case IDC_F1 : sprintf(Temp,"Launch %s - set path in path dialog",m_PathDialog->m_Tool1N); pTTT->lpszText = Temp; break; 
			case IDC_F2 : sprintf(Temp,"Launch %s - set path in path dialog",m_PathDialog->m_Tool2N); pTTT->lpszText = Temp; break; 
			case IDC_F3 : sprintf(Temp,"Launch %s - set path in path dialog",m_PathDialog->m_Tool3N); pTTT->lpszText = Temp; break; 
			case IDC_F4 : sprintf(Temp,"Launch %s - set path in path dialog",m_PathDialog->m_Tool4N); pTTT->lpszText = Temp; break; 
			case IDC_F5 : sprintf(Temp,"Launch %s - set path in path dialog",m_PathDialog->m_Tool5N); pTTT->lpszText = Temp; break; 
			case IDC_OPENFILES : pTTT->lpszText = "Open the DVD folder in explorer"; break; 
			}
			pTTT->hinst = AfxGetResourceHandle();
			return(TRUE);
		}
		else
		{
			switch(nID)
			{
			case IDC_SHRINK : pTTT->lpszText = "Dmarrer DVD Shrink. Choisir le chemin dans le paneau des chemins."; break; 
			case IDC_POWERDVD : pTTT->lpszText = "Dmarrer le lecteur DVD. Choisir le chemin dans le paneau des chemins. Marche avec media player classic, PowerDVD 6.0, et d'autres..."; break; 
			case IDC_RESTORE : pTTT->lpszText = "Annuler le traitement et restaurer les fichiers originaux"; break; 
			case IDC_SHOWADVANCED : pTTT->lpszText = "Ouvrir le panneau avec les options avances"; break; 
			case IDC_SHOWLOW : pTTT->lpszText = "Ouvrir le panneau log avec les messages de MenuShrink"; break; 
			case IDC_SHOWPATH : pTTT->lpszText = "Ouvrir le panneau des chemins pour DVD Shrink, le lecteur DVD et les outils."; break; 
			case IDC_PROCESS : pTTT->lpszText = "Dmarrer le traitement"; break; 
			case IDC_KEEPAUDIO : pTTT->lpszText = "Cocher pour crer des menus statiques avec son. Laisser vierge pour enlever le son et crer des menus plus petits."; break; 
			case IDC_F1 : sprintf(Temp,"Dmarrer %s - Choisir le chemin dans le paneau des chemins",m_PathDialog->m_Tool1N); pTTT->lpszText = Temp; break; 
			case IDC_F2 : sprintf(Temp,"Dmarrer %s - Choisir le chemin dans le paneau des chemins",m_PathDialog->m_Tool2N); pTTT->lpszText = Temp; break; 
			case IDC_F3 : sprintf(Temp,"Dmarrer %s - Choisir le chemin dans le paneau des chemins",m_PathDialog->m_Tool3N); pTTT->lpszText = Temp; break; 
			case IDC_F4 : sprintf(Temp,"Dmarrer %s - Choisir le chemin dans le paneau des chemins",m_PathDialog->m_Tool4N); pTTT->lpszText = Temp; break; 
			case IDC_F5 : sprintf(Temp,"Dmarrer %s - Choisir le chemin dans le paneau des chemins",m_PathDialog->m_Tool5N); pTTT->lpszText = Temp; break; 
			case IDC_OPENFILES : pTTT->lpszText = "Ouvrir le rpertoire du DVD dans explorer"; break; 
			}
			pTTT->hinst = AfxGetResourceHandle();
			return(TRUE);
		}
    }
    return(FALSE);
}

HRESULT CMenuShrinkDlg::ResolveIt(HWND hwnd, LPSTR pszShortcutFile)
{
  HRESULT hres;
  IShellLink* psl;
  char szGotPath[MAX_PATH];
  WIN32_FIND_DATA wfd;

  // Get a pointer to the IShellLink interface.
  hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
                            IID_IShellLink, (void**)&psl);
  if (SUCCEEDED(hres))
  {
    IPersistFile* ppf;

    // Get a pointer to the IPersistFile interface.
    hres = psl->QueryInterface(IID_IPersistFile, (void**)&ppf);
    if (SUCCEEDED(hres))
    {
      WORD wsz[MAX_PATH];

      // Ensure string is Unicode.
      MultiByteToWideChar(CP_ACP, 0, pszShortcutFile, -1, wsz,MAX_PATH);

     // Load the shell link.
     hres = ppf->Load(wsz, STGM_READ);
     if (SUCCEEDED(hres))
     {
       // Resolve the link.
       hres = psl->Resolve(hwnd, SLR_ANY_MATCH);

       if (SUCCEEDED(hres))
       {
          strcpy(szGotPath, pszShortcutFile);
          // Get the path to the link target.
          hres = psl->GetPath(szGotPath, MAX_PATH, (WIN32_FIND_DATA *)&wfd, SLGP_SHORTPATH );
		  strcpy(pszShortcutFile,szGotPath);
        }
      }
      // Release pointer to IPersistFile interface.
      ppf->Release();
    }
    // Release pointer to IShellLink interface.
    psl->Release();
  }
  return hres;
}


HBRUSH CMenuShrinkDlg::OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor) 
{
	HBRUSH hbr = CDialog::OnCtlColor(pDC, pWnd, nCtlColor);
	int id = pWnd->GetDlgCtrlID();
	if(id == IDC_MESSAGE)
	{
		if(m_ReturnedError)
			pDC->SetTextColor(RGB(255, 0, 0)); 
		else
			pDC->SetTextColor(RGB(0, 0, 0)); 
	}
	return hbr;
}

BOOL CMenuShrinkDlg::PreTranslateMessage(MSG* pMsg) 
{
	// This will capture all messages arriving at the app before they're translated and dispatched.
	// We'll just intercept Num Pad message and use them to select which sound to playback...
	if(pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_RETURN)
		return TRUE;
	if(pMsg->message == WM_KEYDOWN && m_Running == 0)
	{
		switch(pMsg->wParam)
		{
			// Must be all CAPS...
		case VK_F1: OnFKey(IDC_F1); break;
		case VK_F2: OnFKey(IDC_F2); break;
		case VK_F3: OnFKey(IDC_F3); break;
		case VK_F4: OnFKey(IDC_F4); break;
		case VK_F5: OnFKey(IDC_F5); break;
		case 'E': ShellExecute(NULL, "open", m_IniFile, "", NULL, SW_SHOWNORMAL); break;
		case 'F': OnOpenfiles(); break;
		case 'O': 
			char Temp[MAX_PATH];
			GetPrivateProfileString("MenuShrink","LastDVD","",Temp,MAX_PATH,m_IniFile);
			if(strcmp(Temp,"") != 0)
				OpenDVD(Temp);
			break;
		default: return CDialog::PreTranslateMessage(pMsg);
		}
		return TRUE;
	}
	else return CDialog::PreTranslateMessage(pMsg);
}

void CMenuShrinkDlg::OnFKey(UINT nID )
{
	char Temp[512];
	// Create path to VIDEO_TS.IFO!
	sprintf(Temp,"\"%s\\VIDEO_TS.IFO\"",ShrinkEngine->m_WorkingDirFullPath);
	char* ToolPath = NULL;
	switch(nID)
	{
	case IDC_F1 : ToolPath = m_PathDialog->m_Tool1; break;
	case IDC_F2 : ToolPath = m_PathDialog->m_Tool2; break;
	case IDC_F3 : ToolPath = m_PathDialog->m_Tool3; break;
	case IDC_F4 : ToolPath = m_PathDialog->m_Tool4; break;
	case IDC_F5 : ToolPath = m_PathDialog->m_Tool5; break;
	}
	if(ToolPath && strcmp(ToolPath,"") != 0 && ShrinkEngine->m_LastVTS != -1)
		ShellExecute(NULL, "open", ToolPath, Temp, NULL, SW_SHOWNORMAL);
}


#include "utils.h"
void CMenuShrinkDlg::OnOpen() 
{
	char Temp[MAX_PATH];
	GetPrivateProfileString("MenuShrink","LastDVD","",Temp,MAX_PATH,m_IniFile);
	char *FName = 0;
	if(strcmp(Temp,"") != 0)
		FName = CatStr(Temp,"/video_ts.ifo");
	if(WindowsMode && FName)
	{
		// Annoying. If there's a '/' in the path, the dialog won't open!!
		for(unsigned int i=0;i<strlen(FName);i++)
			if(FName[i] == '/') FName[i] = '\\';
	}
	static CFileDialog FileDialog(TRUE, CString("*.*"), FName, 0, CString("All Files (*.*)|*.*||"));
	if(CurrentLangage == FRENCH)
		FileDialog.m_ofn.lpstrTitle  = "Ouvrir";
	if (FileDialog.DoModal() == IDOK)
	{
		OpenDVD((char*)(LPCTSTR)FileDialog.GetPathName());
	}
}
