// menushrink.cpp : A small app to shrink down a motion menu cell into a single GOP (possibly
// single I-frame GOP), keeping subpic packs, and dumping audio.
// TODO, TOFIX:
// - The GOP at the first I-frame. Not always closed (except for the first GOP of the cell).
// will need to be fixed if keeping something else than the first I-frame. Also the temporal
// sequence number.
// - I'm not taking into account the PES Header Data length in video frames (always assuming
// the frame starts at 0x17), that might be bad.
// - What about the other various pointers in the navpack DSI and DCI? Adjust?
// - Look for pic header that span two LBAs.

// TODO:
// - In all NavPacks: Set the First/Second/Third ref blocks to 0.
// - In VOBU_SR, for all NavPacks: change the first value (Next VOBU with video) to bfffffff.
// Don't touch the other values if there's audio!
// - In VOBU_SR, change the last value (Previous VOBU with video) to point to the first NavPack.


#include "MenuShrinkEngine.h"
#include <stdlib.h>


// NOTES: v. 400: Playback finally works great in standalone!!!! YEPEEE!

// Flags 
#define KEEP_AUDIO 1
#define JUST_COPY 4
#define PICK_LAST_FRAME 8
#define PICK_LARGEST_FRAME 16
#define PROCESS_ALL_CELLS 0

void DiffFiles(char* FileA, char* FileB);


void Usage()
{
	printf("\n%s\n",VERSION);
	printf("Usage: \tmenushrink VTSM# ...\n");
	printf("OR: \tmenushrink VTSM# flags.\n");
	printf("OR: \tmenushrink VTS_01_0.VOB flags \n");
	printf("OR: \tmenushrink VTS_01_0.VOB flags OUTPUT.VOB \n");
	printf("OR: \tmenushrink VTS_01_0.VOB flags VOBID CELLID \n");
	printf("OR: \tmenushrink VTS_01_0.VOB flags OUTPUT.VOB VOBID CELLID\n");
	printf("\tShrinks all cells in VTS, keeping buttons, and possibly audio.\n");
	printf("\tUpdates the corresponding IFO to reflect changes.\n");
	printf("\tAlso: Finds cell with specified V/CID shrinks and outputs.\n");
	printf("\tflags is a set of flags that works like this.\n");
	printf("\tMenushrink normally keeps the largest I-frame in each cell, however:\n");
	printf("\t\tIf 'a' or 'A' is in flags the audio is preserved\n");
	printf("\t\tIf 'l' or 'L' : the last I-Frame in the cell is used\n");
	printf("\t\tIf 'X' is in flags the specified cell is extracted as is\n");
	printf("\tExample: menushrink 1 a -> shrinks VTS_01_0.VOB keeping the audio.\n");
	printf("\tExample: menushrink VTS_01_0.VOB a  -> same thing.\n");
	printf("\tExample: menushrink VTS_01_0.VOB    -> same thing, no audio.\n");
	printf("\tExample: menushrink VTS_01_0.VOB al -> audio+last frame is picked.\n");
	printf("\tExample: menushrink VTS_01_0.VOB a TEST.VOB -> outputs in TEST.VOB.\n");
	printf("\nCell-by-cell processing:\n");
	printf("\tExample: menushrink VTS_01_0.VOB a 2 2 -> shrinks cell 2/2.\n");
	printf("\tThe created cells are named VCxxx-yy.VOB (xxx=VOBID,yy=CELLID)\n");
	printf("\tExample: menushrink VTS_01_0.VOB a TEST.VOB 2 2 -> outputs in TEST.VOB.\n");
	printf("\tA backup folder is created, and the output replaces VTS_x_0.VOB.\n");
	printf("\tExample: menushrink 1 creates a backup of VTS_01_0.VOB if there \n");
	printf("\tisn't one, keeps the audio and outputs into VTS_01_0.VOB \n");
	printf("\tThe backup is used as the input file the next times around.\n");
	printf("\n\tCoded by jeanl with lots of debugging help from Blutach, thanks man!\n");
	printf("\tcrucial help from r0lZ, and hints and suggestions from mpucoder, jsoto,\n\tand 2COOL...\n");
}

int main(int argc, char* argv[])
{

	if(argc == 4 && strcmp(argv[1],"D") == 0)
	{
		DiffFiles(argv[2],argv[3]);
		return 0;
	}
	if(argc < 2)
	{
		Usage();
		return 0;
	}

	// This is our shrinking engine.
	CMenuShrinkEngine* Shrinker = new CMenuShrinkEngine;

	// Determine input file, based on arguments.
	int WhichVTS = -1;
	if(strstr(argv[1],".VOB") == 0 && strstr(argv[1],".vob") == 0)
	{
		WhichVTS = atoi(argv[1]);
	}

	// Find out processing options.
	uint Flag = PICK_LARGEST_FRAME;
	FrameSelect WhichFrame = FrameSelect_Largest;
	if(argc >= 3)
	{
		if(strchr(argv[2],'a') || strchr(argv[2],'A'))
			Flag |= KEEP_AUDIO;
		if(strchr(argv[2],'X'))
			Flag |= JUST_COPY;
		if(strchr(argv[2],'l') || strchr(argv[2],'L'))
			WhichFrame = FrameSelect_Last;
		if(strchr(argv[2],'f') || strchr(argv[2],'F'))
			WhichFrame = FrameSelect_First;
	}

	// Check we're not stepping on our own feet!
	if(argc == 4 || argc == 6)
	{
		if(strcmp(argv[1],argv[3]) == 0 )
		{
			printf("ERROR! You can't output to the same file you're using for input!\n");
			return -1;
		}
	}

	int OutLBACount = 0;	// LBA we're at in the output file.
	if(argc <= 3)
	{
		// Processing full file, normal mode
		if(WhichVTS != -1)
			OutLBACount = Shrinker->ProcessVTS(WhichVTS,Flag & KEEP_AUDIO, WhichFrame);
		else
			OutLBACount = Shrinker->ProcessVTS(argv[1],Flag & KEEP_AUDIO, WhichFrame);
	}
	else if(argc == 4)
	{
		// Processing full file. Output specified.
		OutLBACount = Shrinker->ShrinkAllCells(argv[1],argv[3],Flag & KEEP_AUDIO, WhichFrame);
	}
	else if(argc == 5)
	{
		// Processing 1 cell no output name given.
		char OutputName[256];
		sprintf(OutputName,"VC%03d-%02d.VOB",atoi(argv[3]),atoi(argv[4]));
		if(Flag & JUST_COPY)
			OutLBACount = Shrinker->ExtractOneCell(argv[1],OutputName,atoi(argv[3]),atoi(argv[4]));
		else
			OutLBACount = Shrinker->ShrinkOneCell(argv[1],OutputName,atoi(argv[3]),atoi(argv[4]),Flag & KEEP_AUDIO, WhichFrame);
	}
	else if(argc == 6)
	{
		// Processing 1 cell.
		if(Flag & JUST_COPY)
			OutLBACount = Shrinker->ExtractOneCell(argv[1],argv[3],atoi(argv[4]),atoi(argv[5]));
		else
			OutLBACount = Shrinker->ShrinkOneCell(argv[1],argv[3],atoi(argv[4]),atoi(argv[5]),Flag & KEEP_AUDIO, WhichFrame);
	}
	else
	{
		Usage();
		return -1;
	}
	// Close, deallocate and exit.
	if(OutLBACount != -1)
		printf("Done... Final size %d kB (%.2f MB)\n",2*OutLBACount, 2.f*OutLBACount/1024);
	return 0;
}


void DiffFiles(char* FileA, char* FileB)
{
	FILE* FDA;
	FILE* FDB;
	FDA = fopen(FileA,"rb");
	FDB = fopen(FileB,"rb");
	if(FDA==0 || FDB == 0)
	{
		printf("Error opening files\n");
		return;
	}

	int LBA = 0;
	int RetA,RetB;
	int ErrCount = 0;
	unsigned char BufferA[SECTOR_SIZE];
	unsigned char BufferB[SECTOR_SIZE];
	while(1)
	{
		// Read A and B.
		RetA = fread(BufferA,1,SECTOR_SIZE,FDA);
		RetB = fread(BufferB,1,SECTOR_SIZE,FDB);
		if(RetA < SECTOR_SIZE || RetB < SECTOR_SIZE)
		{
			if(RetA != RetB)
				printf("File lengths differ! Stopped comparing at LBA %d\n",LBA);
			printf("Done...\n");
			return;
		}

		for(int i=0;i<SECTOR_SIZE;i++)
		{
			if(BufferA[i] != BufferB[i])
			{
				ErrCount++;
				printf("File differ at LBA %d\tbyte 0x%x\t0x%x --\t0x%x\n",LBA,i,
					BufferA[i],BufferB[i]);
				if(ErrCount == 30)
				{
					ErrCount = 0;
					printf("Press enter to continue or CTRL-C to stop...\n");
					getchar();
				}
			}
		}
		LBA++;
	}

}


